/*global escape: false */

/**
 * Functions related to the manipulation and reading of cookies.
 */

var utils = require('./utils');


/**
 * Read cookies currently stored in the browser, returning an object
 * keyed by cookie name.
 *
 * @returns Object
 */

exports.readBrowserCookies = function () {
    if (!utils.isBrowser) {
        throw new Error('readBrowserCookies cannot be called server-side');
    }
    var cookies = {};
    var parts = document.cookie.split(';');
    for (var i = 0, len = parts.length; i < len; i++) {
        var name = parts[i].split('=')[0];
        var value = parts[i].split('=').slice(1).join('=');
        cookies[name] = value;
    }
    return cookies;
};

/**
 * Reads browser cookies and returned the value of the named cookie.
 *
 * @returns {String}
 */

exports.readBrowserCookie = function (name) {
    return exports.readBrowserCookies()[name];
};

/**
 * Creates a string for storing a cookie on the browser.
 *
 * @param {Request Object} req
 * @param {Object} opt
 * @returns {String}
 */

exports.cookieString = function (req, opt) {
    var path = opt.path || utils.getBaseURL(req) + '/';
    var str = escape(opt.name) + '=' + escape(opt.value) + '; path=' + path;
    if (opt.days) {
        var expires = new Date().setTime(
            new Date().getTime() + 1000 * 60 * 60 * 24 * opt.days
        );
        str += '; expires=' + expires.toGMTString();
    }
    return str;
};

/**
 * Sets a cookie on the browser, for use client-side only.
 *
 * @param {Request Object} req
 * @param {Object} opt
 */

exports.setBrowserCookie = function (req, opt) {
    if (!utils.isBrowser) {
        throw new Error('setBrowserCookie cannot be called server-side');
    }
    var str = (typeof opt === 'string') ? opt: exports.cookieString(req, opt);
    //console.log('document.cookie = ' + str);
    document.cookie = str;
};

/**
 * Creates a Set-Cookie header on a response object.
 *
 * @param {Request Object} req
 * @param {Response Object} res
 * @param {Object} opt
 */

exports.setResponseCookie = function (req, res, opt) {
    var str = (typeof opt === 'string') ? opt: exports.cookieString(req, opt);
    if (typeof res !== 'object') {
        res = {code: 200, body: res};
    }
    if (!res.headers) {
        res.headers = {};
    }
    // TODO: is it possible to send multiple set-cookie headers by turning
    // headers into an array like in node?
    // XXX: just replacing all cookies for now - not ideal!
    res.headers['Set-Cookie'] = str;
};
